MfgAide = {}
MfgAide.isMisn = false --true if goods were loaded for the current mission; used to unload goods at end of mission
MfgAide.isWarned = false --true after warning has been shown at least once each misn
MfgAide.destSectors = {}
MfgAide.version = '1.1'

--## Choco ##--

MfgAide.ui, MfgAide.vars = dofile'ui.lua'

local PRINT_COLOR = '\127ffffff'

--print message to chat with correct color
local function print_msg(msg)
	if not type(msg)=="String" then return false end
	print(PRINT_COLOR..msg)
end

--prints warning message if applicable, returns true if warning was shown
local function warning_msg(msg, warn_abort)
	if not type(msg)=="String" then return false end
	
	if not MfgAide.isWarned or not MfgAide.vars.GetVars"IsOneWarning" then
		print_msg('\127ff0000MfgAide WARNING\127o -- '..msg)
		MfgAide.isWarned = true --only show warning once per mission
		
		return true
	elseif MfgAide.vars.GetVars"IsOneWarning" and not MfgAide.vars.GetVars"IsLoadOnWarning" then
		print_msg'Your current setting prevent MfgAide from loading any items. To change them use: \127a0a0a0/mfg options\127o'
		
		return false
	end
end

local function is_mfg_misn()
	local mission_name = GetActiveMissionInfo(1) or '' --TODO compatibility with more than one active mission
	return string.match(mission_name, "^Manufacture") or string.match(mission_name, "^Construct")
end

--Returns table of items to deliver for the active mission
local function get_items()
	local info = {GetActiveMissionInfo(1)}
	MfgAide.destSectors = {}
	
	if info[1] then
		local mission_name = info[1]
		
		local needed_items = {}
		for _,obj in ipairs(info[4] or {}) do
			local num_reqd, item_name, sector, num_delivered = string.match((obj or {})[1] or "", "Deliver (%d+)x? (.+) to (%P+ %u%-%d+)%s*%[?(%d*)/?%d*%]?")
			local needed = (tonumber(num_reqd) or 0) - (tonumber(num_delivered) or 0) --remaining qty needed
			
			if needed>0 then
				if not MfgAide.destSectors[sector] then MfgAide.destSectors[sector]=true end
				
				if not needed_items[item_name] then
					needed_items[item_name]=needed
				else needed_items[item_name] = needed_items[item_name] + needed end --combine multiple entries
			end
		end
		
		return needed_items
	end
	
	return false --no mission data available
end

--Loads as many items as possible for the active mission to the active ship
--options.warn_abort - don't load cargo if not all items available
local function load_items(options)
	local options = options or {}
	local warn_abort = options.warn_abort or not MfgAide.vars.GetVars"IsLoadOnWarning" --use vars setting if not specified
	
	local needed_items = get_items() --qty of items needed by misn, item name as key
	local available_items = {} --qty of items needed from inventory, item name as key
	local load_vol = 0 --volume of goods to be on active ship needed by misn
	local preload_vol --volume of goods already loaded on active ship needed by misn
	
	if not needed_items then print_msg'You must have an active manufacturing mission to use MfgAide'; return end
	if not next(needed_items) and not MfgAide.isMisn then print_msg'This mission has no items to deliver'; return end
	
	--remove any items already loaded in ship from list
	if HasActiveShip() then
		local ship_cargo = GetShipInventory(GetActiveShipID()).cargo
		for _,itemid in ipairs(ship_cargo) do
			local name = GetInventoryItemName(itemid)
			local qty = GetInventoryItemQuantity(itemid) or 0
			local vol = GetInventoryItemVolume(itemid) or 0
			
			if needed_items[name] then
				load_vol = load_vol + math.min(needed_items[name], qty)*vol
				needed_items[name] = math.max(needed_items[name] - qty, 0) --reduce needed qty by qty present
				if needed_items[name]==0 then needed_items[name] = nil end --remove item from list if all qty loaded
			end
		end
	else print_msg'You must have an active ship to use MfgAide'; return end
	preload_vol = load_vol or 0
	
	local stationid = GetStationLocation()
	if stationid then
		local item_list = {} --items in inventory needed by misn
		
		local sta_items = GetStationCargoList(stationid) --inventory at current station
		for _,itemid in ipairs(sta_items) do
			local name = GetInventoryItemName(itemid)
			local qty = GetInventoryItemQuantity(itemid) or 0
			local vol = GetInventoryItemVolume(itemid) or 0
			
			if needed_items[name] then
				local item = {itemid=itemid, name=name, qty=math.min(qty, needed_items[name]), vol=vol}
				table.insert(item_list, item)
				
				available_items[name] = (available_items[name] or 0) + qty
				
				needed_items[name] = math.max(needed_items[name] - qty, 0) --reduce needed qty by inventory qty
				if needed_items[name]==0 then needed_items[name] = nil end --remove item from list if all qty in inventory
			end
		end
		
		table.sort(item_list, function(a,b) return tonumber(a.vol)>tonumber(b.vol) end) --sort item_list, largest volume first
		
		--identify any needed items that are not available
		local is_missing = false --true if player doesn't have at least one required item
		local missing_items = 'The following items are not available in inventory: '
		for name,num in pairs(needed_items) do
			missing_items = missing_items.." "..num.."x "..name..","
			is_missing = true
		end
		
		local cancel_load_str = warn_abort and ' (loading canceled)' or ''
		
		--warning for not all items available
		if is_missing then
			warning_msg(missing_items..cancel_load_str)
			if warn_abort then return false end --don't load cargo if warning message
		end
		
		local max_cargo = GetActiveShipMaxCargo()
		local free_space = max_cargo - GetActiveShipCargoCount()
		local largest_vol = item_list[1] and item_list[1].vol or 0
		
		--warning for not enough free space for largest item
		if largest_vol > free_space + preload_vol then
			warning_msg('There is not enough free space in your current ship to load one or more item(s): '..largest_vol..' cu required'..cancel_load_str)
			if warn_abort then return false end --don't load cargo if warning message
		end
		
		local load_cargo = {} --list of items to be loaded at this time
		for _,item in ipairs(item_list) do
			local move_qty = math.min(math.floor(free_space/item.vol), item.qty)
			if move_qty>0 then
				table.insert(load_cargo, {itemid=item.itemid, quantity=move_qty})
				free_space = free_space - move_qty*item.vol
				load_vol = load_vol + move_qty*item.vol
				
				if free_space <=0 then break end --TODO any 0cu cargo exists???
			end
		end
		
		if load_vol>0 then --TODO use next instead
			LoadCargo(load_cargo) --TODO cancel dlg?
		end
		
		return true
	else print_msg'You must be docked at a station to use MfgAide'; return end
end

--checks misn active, sector and auto-loading enabled then loads items
local function tryAutoload()
	local needed_items = get_items() --generates MfgAide.destSectors list
	
	if not is_mfg_misn() or not next(needed_items or {}) then --[[console_print'no objectives found';]] return end --no items to deliver on this mission
	
	local stationid = GetStationLocation()
	local sector_name = stationid and ShortLocationStr(GetSectorIDOfStation(stationid))
	if MfgAide.vars.GetVars"isLoadDestOnly" and (not sector_name or not MfgAide.destSectors[sector_name]) then --[[console_print'dest does not match';]] return end --not at destination
	
	if MfgAide.vars.GetVars"IsAutoLoad" then MfgAide.isMisn = load_items{warn_abort=true} or MfgAide.isMisn end --do auto-load
end

function MfgAide:PLAYER_ENTERED_GAME(event, data)
	RegisterEvent(MfgAide, "SHOW_STATION")
	RegisterEvent(MfgAide, "HUD_SHOW")
end

function MfgAide:SHOW_STATION(event, data)
	UnregisterEvent(MfgAide, "SHOW_STATION")
	UnregisterEvent(MfgAide, "HUD_SHOW")
	RegisterEvent(MfgAide, "ENTERED_STATION")
	
	tryAutoload()
end

function MfgAide:HUD_SHOW(event, data)
	UnregisterEvent(MfgAide, "SHOW_STATION")
	UnregisterEvent(MfgAide, "HUD_SHOW")
	RegisterEvent(MfgAide, "ENTERED_STATION")
end

function MfgAide:MISSION_ADDED(event, data)
	local info = {GetActiveMissionInfo(1)}
	
	RegisterEvent(MfgAide, "MISSION_UPDATED") --wait until mission updates to continue
end

--Check that all required items are available when mission begins
function MfgAide:MISSION_UPDATED(event, data)
	UnregisterEvent(MfgAide, "MISSION_UPDATED")
	
	tryAutoload()
end

--Unload all cargo when misn completes
function MfgAide:MISSION_REMOVED(event, data)
	if MfgAide.isMisn then --only do unload if items were loaded this mission
		--unload all cargo --TODO find just the mfg item and unload it?
		if HasActiveShip() and MfgAide.vars.GetVars"IsAutoUnload" then --do auto-unload
			local unload_items = {}
			
			local ship_cargo = GetShipInventory(GetActiveShipID()).cargo
			for _,itemid in ipairs(ship_cargo) do
				local qty = GetInventoryItemQuantity(itemid) or 0
				table.insert(unload_items, {itemid=itemid, quantity=qty})
			end
			CheckStorageAndUnloadCargo(unload_items)
		end
		
		MfgAide.isMisn = false
		MfgAide.isWarned = false
	end
end

function MfgAide:ENTERED_STATION(event, data)
	tryAutoload()
end

function MfgAide:PLAYER_LOGGED_OUT(event, data)
	MfgAide.isWarned = false
	
	UnregisterEvent(MfgAide, "ENTERED_STATION")
	UnregisterEvent(MfgAide, "MISSION_UPDATED")
end

MfgAide.Inputs = {load=true,options=true,help=true} --list of valid commands
function MfgAide:input(args)
	args = args or {}
	local command = args[1] and args[1]:lower()
	if command then table.remove(args, 1) end
	
	if MfgAide.Inputs[command] then
		MfgAide[command](MfgAide, args)
	else MfgAide.load(MfgAide) end
end

function MfgAide:load(args) MfgAide.isMisn = load_items() or MfgAide.isMisn end
function MfgAide:options(args) MfgAide.ui:Show() end
function MfgAide:help(args) MfgAide.ui:ShowHelp() end

RegisterEvent(MfgAide, "PLAYER_ENTERED_GAME")
RegisterEvent(MfgAide, "MISSION_ADDED")
RegisterEvent(MfgAide, "MISSION_REMOVED")
RegisterEvent(MfgAide, "PLAYER_LOGGED_OUT")
RegisterUserCommand("mfg", MfgAide.input)